﻿using System;
using System.Drawing;
using System.Windows.Forms;
using System.IO.Ports;

namespace CM_2408C2R_CS
{
    public partial class Form1 : Form
    {
        /* Command ********************************************************** */
        private const byte CMD_ReadData   = 0x81;               // 命令: 讀取資料
        private const byte CMD_L1_Set     = 0x91;               // 命令: L1設定值
        private const byte CMD_L2_Set     = 0x92;               // 命令: L2設定值
        private const byte CMD_L1_Output  = 0xA1;               // 命令: L1輸出值
        private const byte CMD_L2_Output  = 0xA2;               // 命令: L2輸出值
        private const byte CMD_L1_OnOff   = 0xC1;               // 命令: L1 ON/OFF
        private const byte CMD_L2_OnOff   = 0xC2;               // 命令: L2 ON/OFF
        private const byte CMD_L1_TrigOut = 0xD1;               // 命令: L1觸發輸出延遲
        private const byte CMD_L2_TrigOut = 0xD2;               // 命令: L2觸發輸出延遲

        /* Lx_Set =========================================================== */
                                                                // bit7~4 最大電流
        private const byte SET_ExtOnOff = 0x02;                 // bit1   外部ON/OFF: 0=除能, 1=致能
        private const byte SET_OnOff    = 0x01;                 // bit0   ON/OFF    : 0=OFF , 1=ON

        /* ACK ============================================================== */
        private const byte ACK_Failed  = 0x80;                  // bit7 命令執行狀態: 0=成功, 1=失敗
        private const byte ACK_LowVolt = 0x10;                  // bit4 電源低電壓  : 0=No,   1=Yes
        private const byte ACK_L2_OC   = 0x02;                  // bit1 L2過流      : 0=No,   1=Yes
        private const byte ACK_L1_OC   = 0x01;                  // bit0 L1過流      : 0=No,   1=Yes

        /* Baud Rate ======================================================== */
        private const byte BaudRate_LS = 0;                     // 9600 8-N-1
        private const byte BaudRate_MS = 1;                     // 38400 8-N-1
        private const byte BaudRate_HS = 2;                     // 115200 8-N-1

        /* ================================================================== */
        private const ushort LineMax = 256;                     // 最大行數

        /* ****************************************************************** */
        private static ushort LineNo = 0;                       // 行數編號
        private static ushort LineCount = 0;                    // 行數計數
        /* ------------------------------------------------------------------ */
        private static String[] PortNo;                         // 埠編號
        private static bool FG_ComPort = false;                 // 旗標-開啟串列埠
        private static bool FG_ReadData = false;                // 旗標-讀取資料
        /* ------------------------------------------------------------------ */
        private static byte CNT_AckTimeOut = 0;                 // 計時-回應逾時
        private static byte CNT_AskAuto = 0;                    // 計時-自動詢問
        private static byte CNT_Flash = 0;                      // 計時-閃爍
        /* 最大電流 --------------------------------------------------------- */
        // 2A版本
        private static int[] MaxCurrent = new int[] { 500, 600, 700, 800, 900, 1000, 1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800, 1900, 2000 };
        //private static int[] OutputCurrent = new int[201];

        /* TX =============================================================== */
        private static bool FG_Transmit = false;                // 旗標-傳送
        private static byte[] TX_Packet = new byte[7];          // 傳送-封包
        private static byte TX_Command = 0;                     // 傳送-命令
        private static byte TX_Count = 0;                       // 傳送-計次

        /* RX =============================================================== */
        private static bool FG_Receive = false;                 // 旗標-接收
        private static byte[] RX_Packet = new byte[13];         // 接收-封包
        private static byte RX_Index = 0;                       // 接收-封包索引
        private static byte RX_Length = 0;                      // 接收-封包長度
        private static byte ACK = 0;                            // 回應碼

        /* L1 =============================================================== */
        private static bool CHG_L1_Set = false;                 // 變更-L1設定值
        private static bool CHG_L1_Output = false;              // 變更-L1輸出值
        private static bool CHG_L1_OnOff = false;               // 變更-L1 ON/OFF
        private static bool CHG_L1_TrigOut = false;             // 變更-L1觸發輸出延遲
        private static byte L1_Set = 0;                         // L1設定值
        private static int L1_Output = 0;                       // L1輸出值
        private static byte L1_TrigOut = 0;                     // L1觸發輸出延遲

        /* L2 =============================================================== */
        private static bool CHG_L2_Set = false;                 // 變更-L2設定值
        private static bool CHG_L2_Output = false;              // 變更-L2輸出值
        private static bool CHG_L2_OnOff = false;               // 變更-L2 ON/OFF
        private static bool CHG_L2_TrigOut = false;             // 變更-L2觸發輸出延遲
        private static byte L2_Set = 0;                         // L2設定值
        private static int L2_Output = 0;                       // L2輸出值
        private static byte L2_TrigOut = 0;                     // L2觸發輸出延遲

        /* ****************************************************************** */

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // 產生200階電流(新版本為數值直接對應，故註解)
            //for (int i = 0; i < OutputCurrent.Length; i++)
            //    OutputCurrent[i] = i * 4;

            cbb_L1_MaxC.SelectedIndex = 0;                      // L1最大電流 = 0.5A
            lb_L1_MaxC.Text = (MaxCurrent[0] / 1000F).ToString("0.0");

            cbb_L2_MaxC.SelectedIndex = 0;                      // L2最大電流 = 0.5A
            lb_L2_MaxC.Text = (MaxCurrent[0] / 1000F).ToString("0.0");

            cbb_BaudRate.SelectedIndex = BaudRate_HS;           // 115200 8-N-1
            ComPort_Detect();                                   // 串列埠檢知
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            if (FG_ComPort == true)                             // 串列埠已開啟?
            {
                if (FG_Receive == false)                        // 尚無接收有效封包?
                {
                    if (FG_Transmit == false)                   // 尚無傳送命令?
                    {
                        Check_TX_Command();                     // 檢查-傳送命令
                    }
                    else                                        // 已有傳送命令
                    {
                        Check_RX_TimeOut();                     // 檢查-接收逾時
                    }
                    Ask_Auto();                                 // 自動詢問
                }
                else                                            // 已完成接收封包
                {
                    FG_Receive = false;
                    CNT_AckTimeOut = 0;
                    FG_Transmit = false;
                    TX_Count = 0;
                    if (FG_ReadData == false)                   // 未讀取過資料?
                    {
                        Dsp_RX_Packet(Color.White, 13);
                        Update_L1();                            // 更新-L1參數
                        Update_L2();                            // 更新-L2參數
                        Check_ACK();                            // 檢查-回應碼
                        FG_ReadData = true;
                    }
                    else                                        // 已讀取過資料
                    {
                        Dsp_RX_Packet(Color.White, 5);
                        Check_ACK();                            // 檢查-回應碼
                    }
                }

                if ((ACK & 0x13) != 0)                          // 電源低電壓/L2過流/L1過流 ?
                {
                    Alarm_Flash();                              // 警告訊息-閃爍 (2Hz)
                }
            }
        }

        /* ================================================================== */

        private void Check_TX_Command()                         // 檢查-傳送命令
        {
            if (CHG_L1_Set == true)                             // 變更-L1設定值?
            {
                TX_L1_Set(); return;                            // 傳送-L1設定值
            }

            if (CHG_L2_Set == true)                             // 變更-L2設定值?
            {
                TX_L2_Set(); return;                            // 傳送-L2設定值
            }

            if (CHG_L1_Output == true)                          // 變更-L1輸出值?
            {
                TX_L1_Output(); return;                         // 傳送-L1輸出值
            }

            if (CHG_L2_Output == true)                          // 變更-L2輸出值?
            {
                TX_L2_Output(); return;                         // 傳送-L2輸出值
            }

            if (CHG_L1_OnOff == true)                           // 變更-L1 ON/OFF ?
            {
                TX_L1_OnOff(); return;                          // 傳送-L1 ON/OFF
            }

            if (CHG_L2_OnOff == true)                           // 變更-L2 ON/OFF ?
            {
                TX_L2_OnOff(); return;                          // 傳送-L2 ON/OFF
            }

            if (CHG_L1_TrigOut == true)                         // 變更-L1觸發輸出延遲?
            {
                TX_L1_TrigOut(); return;                        // 傳送-L1觸發輸出延遲
            }

            if (CHG_L2_TrigOut == true)                         // 變更-L2觸發輸出延遲?
            {
                TX_L2_TrigOut(); return;                        // 傳送-L2觸發輸出延遲
            }
        }

        private void Check_RX_TimeOut()                         // 檢查-接收逾時
        {
            if (++CNT_AckTimeOut >= 10)                         // 回應逾時? (10 * 50ms = 500ms)
            {
                CNT_AckTimeOut = 0;
                if (TX_Count > 2)
                {
                    Dsp_String(Color.Red, " 逾時未回應!");
                    ComPort_Disconnect();                       // 串列埠離線
                }
                else
                {
                    switch (TX_Command)
                    {
                        case CMD_ReadData  : TX_ReadData(); break;    // 傳送-讀取資料
                        case CMD_L1_Set    : TX_L1_Set(); break;      // 傳送-L1設定值
                        case CMD_L2_Set    : TX_L2_Set(); break;      // 傳送-L2設定值
                        case CMD_L1_Output : TX_L1_Output(); break;   // 傳送-L1輸出值
                        case CMD_L2_Output : TX_L2_Output(); break;   // 傳送-L2輸出值
                        case CMD_L1_OnOff  : TX_L1_OnOff(); break;    // 傳送-L1 ON/OFF
                        case CMD_L2_OnOff  : TX_L2_OnOff(); break;    // 傳送-L2 ON/OFF
                        case CMD_L1_TrigOut: TX_L1_TrigOut(); break;  // 傳送-L1觸發輸出延遲
                        case CMD_L2_TrigOut: TX_L2_TrigOut(); break;  // 傳送-L2觸發輸出延遲
                    }
                }
            }
        }

        private void Ask_Auto()                                 // 自動詢問
        {
            if (cbb_L1_MaxC.DroppedDown == true) { CNT_AskAuto = 0; return; }
            if (cbb_L2_MaxC.DroppedDown == true) { CNT_AskAuto = 0; return; }

            if (nud_L1_Output.Focused == true) { CNT_AskAuto = 0; return; }
            if (nud_L2_Output.Focused == true) { CNT_AskAuto = 0; return; }

            if (nud_L1_TrigOut.Focused == true) { CNT_AskAuto = 0; return; }
            if (nud_L2_TrigOut.Focused == true) { CNT_AskAuto = 0; return; }

            if (++CNT_AskAuto >= 60)                            // 自動詢問? (60 * 50ms = 3sec)
            {
                TX_ReadData();                                  // 傳送-讀取資料
            }
        }

        private void Update_L1()                                // 更新-L1參數
        {
            L1_Set = (byte)(RX_Packet[4] & 0xF3);

            if ((L1_Set & SET_OnOff) == 0)                      // L1 OFF ?
            {
                lb_L1_LED.BackColor = Color.DimGray;
                btn_L1_OnOff.Text = "ON";
            }
            else                                                // L1 ON
            {
                lb_L1_LED.BackColor = Color.Green;
                btn_L1_OnOff.Text = "OFF";
            }

            if ((L1_Set & SET_ExtOnOff) == 0)                   // L1外部ON/OFF-除能?
            {
                ckb_L1_ExtOnOff.Checked = false;
                lb_L1_LED.Visible = true;
                btn_L1_OnOff.Visible = true;
            }
            else                                                // L1外部ON/OFF-致能
            {
                ckb_L1_ExtOnOff.Checked = true;
                lb_L1_LED.Visible = false;
                btn_L1_OnOff.Visible = false;
            }

            /* L1最大電流 --------------------------------------------------- */
            cbb_L1_MaxC.SelectedIndex = (L1_Set >> 4) & 0x0F;
            int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];

            lb_L1_MaxC.Text = (MaxC / 1000F).ToString("0.0");   // 以小數點顯示
            //tkb_L1_Output.TickFrequency = MaxC / 10;
            nud_L1_Output.Maximum = (decimal)(MaxC / 1000F);    // 以小數點顯示
            tkb_L1_Output.Maximum = MaxC;

            /* L1輸出值 ----------------------------------------------------- */
            int Output = RX_Packet[5];
            Output <<= 8;
            Output |= RX_Packet[6];
            //Output = (int)(Output * 2.5);   // 2.5倍率顯示(新版直接對應，故註解)

            nud_L1_Output.Value = (decimal)(Output / 1000F);   // 以小數點顯示
            tkb_L1_Output.Value = Output;
            L1_Output = Output;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(Output * 100) / MaxC);

            /* L1觸發輸出延遲 ----------------------------------------------- */
            L1_TrigOut = RX_Packet[7];
            nud_L1_TrigOut.Value = L1_TrigOut;
            lb_L1_TrigOut.Text = Convert.ToString(L1_TrigOut * 5) + "ms";
        }

        private void Update_L2()                                // 更新-L2參數
        {
            L2_Set = (byte)(RX_Packet[8] & 0xF3);

            if ((L2_Set & SET_OnOff) == 0)                      // L2 OFF ?
            {
                lb_L2_LED.BackColor = Color.DimGray;
                btn_L2_OnOff.Text = "ON";
            }
            else                                                // L2 ON
            {
                lb_L2_LED.BackColor = Color.Green;
                btn_L2_OnOff.Text = "OFF";
            }

            if ((L2_Set & SET_ExtOnOff) == 0)                   // L2外部ON/OFF-除能?
            {
                ckb_L2_ExtOnOff.Checked = false;
                lb_L2_LED.Visible = true;
                btn_L2_OnOff.Visible = true;
            }
            else                                                // L2外部ON/OFF-致能
            {
                ckb_L2_ExtOnOff.Checked = true;
                lb_L2_LED.Visible = false;
                btn_L2_OnOff.Visible = false;
            }

            /* L2最大電流 --------------------------------------------------- */
            cbb_L2_MaxC.SelectedIndex = (L2_Set >> 4) & 0x0F;
            int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];

            lb_L2_MaxC.Text = (MaxC / 1000F).ToString("0.0");   // 以小數點顯示
            //tkb_L2_Output.TickFrequency = MaxC / 10;
            nud_L2_Output.Maximum = (decimal)(MaxC / 1000F);    // 以小數點顯示
            tkb_L2_Output.Maximum = MaxC;

            /* L2輸出值 ----------------------------------------------------- */
            int Output = RX_Packet[9];
            Output <<= 8;
            Output |= RX_Packet[10];
            //Output = (int)(Output * 2.5);   // 2.5倍率顯示(新版直接對應，故註解)

            nud_L2_Output.Value = (decimal)(Output / 1000F);    // 以小數點顯示
            tkb_L2_Output.Value = Output;
            L2_Output = Output;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(Output * 100) / MaxC);

            /* L2觸發輸出延遲 ----------------------------------------------- */
            L2_TrigOut = RX_Packet[11];
            nud_L2_TrigOut.Value = L2_TrigOut;
            lb_L2_TrigOut.Text = Convert.ToString(L2_TrigOut * 5) + "ms";
        }

        private void Check_ACK()                                // 檢查-回應碼
        {
            ACK = (byte)(RX_Packet[3] & 0x93);

            if ((ACK & ACK_Failed) == ACK_Failed)               // bit7: 命令執行失敗?
            {
                Dsp_String(Color.Red, " 命令執行失敗!");
            }

            if ((ACK & ACK_LowVolt) == ACK_LowVolt)             // bit4: 電源低電壓?
            {
                Dsp_String(Color.Red, " 電源低電壓!");
            }
            else lb_LowVolt.BackColor = Color.DimGray;

            if ((ACK & 0x03) == 0x03)                           // L2過流 & L1過流 ?
            {
                Dsp_String(Color.Red, " L1-L2過流!");
            }
            else
            {
                if ((ACK & ACK_L2_OC) == ACK_L2_OC)             // bit1: L2過流?
                {
                    Dsp_String(Color.Red, " L2過流!");
                }
                else lb_L2_OC.BackColor = Color.DimGray;

                if ((ACK & ACK_L1_OC) == ACK_L1_OC)             // bit0: L1過流?
                {
                    Dsp_String(Color.Red, " L1過流!");
                }
                else lb_L1_OC.BackColor = Color.DimGray;
            }

            if ((ACK & 0x13) != 0)                              // 電源低電壓/L2過流/L1過流 ?
            {
                Lx_GroupBox_Enabled(false);                     // Lx群組方塊 (不啟用)
            }
            else
            {
                Lx_GroupBox_Enabled(true);                      // Lx群組方塊 (啟用)
            }
        }

        private void Alarm_Flash()                              // 警告訊息-閃爍 (2Hz)
        {
            if ((ACK & ACK_LowVolt) == ACK_LowVolt)             // bit4: 電源低電壓?
            {
                if (CNT_Flash == 0) lb_LowVolt.BackColor = Color.Red;
                if (CNT_Flash == 5) lb_LowVolt.BackColor = Color.DimGray;
            }

            if ((ACK & ACK_L2_OC) == ACK_L2_OC)                 // bit1: L2過流?
            {
                if (CNT_Flash == 0) lb_L2_OC.BackColor = Color.Red;
                if (CNT_Flash == 5) lb_L2_OC.BackColor = Color.DimGray;
            }

            if ((ACK & ACK_L1_OC) == ACK_L1_OC)                 // bit0: L1過流?
            {
                if (CNT_Flash == 0) lb_L1_OC.BackColor = Color.Red;
                if (CNT_Flash == 5) lb_L1_OC.BackColor = Color.DimGray;
            }

            if (++CNT_Flash > 9) CNT_Flash = 0;                 // 10 * 50ms = 500ms
        }

        /* RichTextBox 顯示 ************************************************* */

        private void Dsp_String(Color text_color, string str)   // 顯示字串
        {
            if (++LineCount > LineMax)                          // 行數計數 > 最大行數?
            {
                LineCount = 1;
                richTextBox.Clear();
            }

            if (LineCount > 1) richTextBox.AppendText("\n");
            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();

            /* 行數編號顯示 ------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.DimGray;
            richTextBox.SelectionColor = Color.White;
            richTextBox.AppendText(String.Format("{0:X4}", LineNo++));

            /* 字串顯示 ----------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.Black;
            richTextBox.SelectionColor = text_color;
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        private void Dsp_TX_Packet(Color text_color, byte length)  // 顯示 TX 封包
        {
            if (++LineCount > LineMax)                          // 行數計數 > 最大行數?
            {
                LineCount = 1;
                richTextBox.Clear();
            }

            if (LineCount > 1) richTextBox.AppendText("\n");
            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();

            /* 行數編號顯示 ------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.DimGray;
            richTextBox.SelectionColor = Color.White;
            richTextBox.AppendText(String.Format("{0:X4}", LineNo++));

            /* TX 封包 -> 字串 ---------------------------------------------- */
            string str = " TX:";
            byte i;
            for (i = 0; i < length; i++) str += String.Format(" {0:X2}", TX_Packet[i]);
            str += " (" + Convert.ToString(TX_Count) + ")";

            /* 字串顯示 ----------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.Black;
            richTextBox.SelectionColor = text_color;
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        private void Dsp_RX_Packet(Color text_color, byte length)  // 顯示 RX 封包
        {
            if (++LineCount > LineMax)                          // 行數計數 > 最大行數?
            {
                LineCount = 1;
                richTextBox.Clear();
            }

            if (LineCount > 1) richTextBox.AppendText("\n");
            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();

            /* 行數編號顯示 ------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.DimGray;
            richTextBox.SelectionColor = Color.White;
            richTextBox.AppendText(String.Format("{0:X4}", LineNo++));

            /* RX 封包 -> 字串 ---------------------------------------------- */
            string str = " RX:";
            byte i;
            for (i = 0; i < length; i++) str += String.Format(" {0:X2}", RX_Packet[i]);

            /* 字串顯示 ----------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.Black;
            richTextBox.SelectionColor = text_color;
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        /* 串列埠 *********************************************************** */

        private void cbb_ComPort_Click(object sender, EventArgs e)
        {
            ComPort_Detect();                                   // 串列埠檢知
        }

        private void btn_Connect_Click(object sender, EventArgs e)
        {
            bool state = false;

            if (serialPort.IsOpen)                              // 串列埠已開啟?
            {
                ComPort_Disconnect();                           // 串列埠離線
            }
            else                                                // 串列埠未開啟
            {
                PortNo = SerialPort.GetPortNames();
                if (PortNo.Length >= 1)
                {
                    if (cbb_ComPort.Text == "")
                    {
                        MessageBox.Show(this, "請選擇有效存在的串列埠.", "未選擇串列埠", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    }
                    else
                    {
                        foreach (string port in PortNo)
                        {
                            if (port == cbb_ComPort.Text)
                            {
                                state = true;
                                break;
                            }
                        }
                        if (state == true)
                        {
                            if (serialPort.IsOpen)              // 串列埠已開啟?
                            {
                                ComPort_Disconnect();           // 串列埠離線
                            }
                            else                                // 串列埠未開啟
                            {
                                ComPort_Connect();              // 串列埠連線
                                if (FG_ComPort == true)         // 串列埠已連線?
                                {
                                    TX_ReadData();              // 傳送-讀取資料
                                }
                            }
                        }
                    }
                }
                else
                {
                    cbb_ComPort.Items.Clear();
                    MessageBox.Show(this, "請檢視串列埠.", "未發現任何串列埠", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            }
        }

        private void ComPort_Detect()                           // 串列埠檢知
        {
            string Port_No = cbb_ComPort.Text;
            cbb_ComPort.Text = "";
            cbb_ComPort.Items.Clear();

            PortNo = SerialPort.GetPortNames();                 // 取得存在的串列埠
            if (PortNo.Length >= 1)
            {
                foreach (string port in PortNo)                 // 將找到的現有串列埠加入下拉式方塊中的項目
                {
                    cbb_ComPort.Items.Add(port);
                    if (port == Port_No) cbb_ComPort.Text = Port_No;
                }

                if (cbb_ComPort.Text == "")
                {
                    cbb_ComPort.Text = PortNo[0];               // 下拉式方塊先顯示一個現存的串列埠
                }
            }
            else
            {
                MessageBox.Show(this, "請檢視串列埠.", "未發現任何串列埠", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
        }

        private void ComPort_Connect()                          // 串列埠連線
        {
            /* 設定串列埠鮑率 ----------------------------------------------- */
            serialPort.PortName = cbb_ComPort.Text;
            switch (cbb_BaudRate.SelectedIndex)
            {
                case BaudRate_LS: serialPort.BaudRate = 9600; break;    // 9600 8-N-1
                case BaudRate_MS: serialPort.BaudRate = 38400; break;   // 38400 8-N-1
                case BaudRate_HS: serialPort.BaudRate = 115200; break;  // 115200 8-N-1
            }

            try
            {
                serialPort.Open();
                serialPort.DiscardInBuffer();
                serialPort.DiscardOutBuffer();

                cbb_BaudRate.Enabled = false;
                cbb_ComPort.Enabled = false;

                btn_Connect.Text = "Disconnect";
                tss_Status.Text = "Connect";
                FG_ComPort = true;
            }
            catch (Exception)
            {
                MessageBox.Show(this, "串列埠開啟程序發生不可預期錯誤!\r串列埠開啟失敗.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void ComPort_Disconnect()                       // 串列埠離線
        {
            FG_ComPort = false;
            FG_Receive = false;
            FG_Transmit = false;
            FG_ReadData = false;

            ACK = 0;
            RX_Index = 0;
            TX_Command = 0;
            TX_Count = 0;

            CHG_L1_Set = false;
            CHG_L1_Output = false;
            CHG_L1_OnOff = false;
            CHG_L1_TrigOut = false;

            CHG_L2_Set = false;
            CHG_L2_Output = false;
            CHG_L2_OnOff = false;
            CHG_L2_TrigOut = false;

            try
            {
                serialPort.Close();
            }
            catch (Exception)
            {
                ComPort_Detect();                               // 串列埠檢知
            }

            Lx_GroupBox_Enabled(false);                         // Lx群組方塊 (不啟用)
            lb_L1_LED.BackColor = Color.DimGray;
            lb_L2_LED.BackColor = Color.DimGray;

            lb_L1_OC.BackColor = Color.DimGray;
            lb_L2_OC.BackColor = Color.DimGray;
            lb_LowVolt.BackColor = Color.DimGray;

            cbb_BaudRate.Enabled = true;
            cbb_ComPort.Enabled = true;

            btn_Connect.Text = "Connect";
            tss_Status.Text = "Disconnect";
        }

        private void Lx_GroupBox_Enabled(bool s)                // Lx群組方塊啟用否?
        {
            gpb_L1_OnOffCtrl.Enabled = s;
            gpb_L1_Output.Enabled = s;
            gpb_L1_TrigOut.Enabled = s;

            gpb_L2_OnOffCtrl.Enabled = s;
            gpb_L2_Output.Enabled = s;
            gpb_L2_TrigOut.Enabled = s;
        }

        /* 串列埠傳送 ******************************************************* */

        private void TX_ReadData()                              // 傳送-讀取資料
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 5;                               // 封包長度
                TX_Packet[3] = CMD_ReadData;                    // 命令碼

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 4; i++) checksum += TX_Packet[i];
                TX_Packet[4] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_ReadData;
                TX_Count = 0;
                FG_Transmit = true;
                FG_ReadData = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.White, 5);
            else Dsp_TX_Packet(Color.Magenta, 5);

            try
            {
                serialPort.Write(TX_Packet, 0, 5);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-讀取資料] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L1_Set()                                // 傳送-L1設定值
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L1_Set;                      // 命令碼
                TX_Packet[4] = L1_Set;                          // L1設定值

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L1_Set;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L1_Set = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Yellow, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L1設定值] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L2_Set()                                // 傳送-L2設定值
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L2_Set;                      // 命令碼
                TX_Packet[4] = L2_Set;                          // L2設定值

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L2_Set;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L2_Set = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Yellow, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L2設定值] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L1_Output()                             // 傳送-L1輸出值
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 7;                               // 封包長度
                TX_Packet[3] = CMD_L1_Output;                   // 命令碼
                TX_Packet[4] = (byte)(L1_Output >> 8);          // L1輸出值 (HB)
                TX_Packet[5] = (byte)L1_Output;                 // L1輸出值 (LB)

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 6; i++) checksum += TX_Packet[i];
                TX_Packet[6] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L1_Output;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L1_Output = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Cyan, 7);
            else Dsp_TX_Packet(Color.Magenta, 7);

            try
            {
                serialPort.Write(TX_Packet, 0, 7);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L1輸出值] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L2_Output()                             // 傳送-L2輸出值
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 7;                               // 封包長度
                TX_Packet[3] = TX_Command = CMD_L2_Output;      // 命令碼
                TX_Packet[4] = (byte)(L2_Output >> 8);          // L2輸出值 (HB)
                TX_Packet[5] = (byte)L2_Output;                 // L2輸出值 (LB)
                //TX_Packet[6] = (byte)(0xA7 + TX_Command + TX_Packet[4] + TX_Packet[5]);  // 檢查碼

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 6; i++) checksum += TX_Packet[i];
                TX_Packet[6] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Count = 0;
                FG_Transmit = true;
                CHG_L2_Output = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Cyan, 7);
            else Dsp_TX_Packet(Color.Magenta, 7);

            try
            {
                serialPort.Write(TX_Packet, 0, 7);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L2輸出值] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L1_OnOff()                              // 傳送-L1 ON/OFF
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L1_OnOff;                    // 命令碼
                TX_Packet[4] = (byte)(L1_Set & SET_OnOff);      // L1 ON/OFF

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L1_OnOff;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L1_OnOff = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Lime, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L1 ON/OFF] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L2_OnOff()                              // 傳送-L2 ON/OFF
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L2_OnOff;                    // 命令碼
                TX_Packet[4] = (byte)(L2_Set & SET_OnOff);      // L2 ON/OFF

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L2_OnOff;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L2_OnOff = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Lime, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L2 ON/OFF] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L1_TrigOut()                            // 傳送-L1觸發輸出延遲
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L1_TrigOut;                  // 命令碼
                TX_Packet[4] = L1_TrigOut;                      // L1觸發輸出延遲

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L1_TrigOut;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L1_TrigOut = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Pink, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L1觸發輸出延遲] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        private void TX_L2_TrigOut()                            // 傳送-L2觸發輸出延遲
        {
            if (FG_Transmit == false)
            {
                TX_Packet[0] = 0x4D;                            // Header 'M'
                TX_Packet[1] = 0x53;                            // Header 'S'
                TX_Packet[2] = 6;                               // 封包長度
                TX_Packet[3] = CMD_L2_TrigOut;                  // 命令碼
                TX_Packet[4] = L2_TrigOut;                      // L2觸發輸出延遲

                ushort checksum = 0;
                byte i;
                for (i = 0; i < 5; i++) checksum += TX_Packet[i];
                TX_Packet[5] = (byte)(checksum & 0xFF);         // 檢查碼

                TX_Command = CMD_L2_TrigOut;
                TX_Count = 0;
                FG_Transmit = true;
                CHG_L2_TrigOut = false;
            }

            CNT_AckTimeOut = 0;
            CNT_AskAuto = 0;
            if (++TX_Count == 1) Dsp_TX_Packet(Color.Pink, 6);
            else Dsp_TX_Packet(Color.Magenta, 6);

            try
            {
                serialPort.Write(TX_Packet, 0, 6);
            }
            catch (Exception)
            {
                ComPort_Disconnect();                           // 串列埠離線
                MessageBox.Show(this, "[傳送-L2觸發輸出延遲] 發生不可預期錯誤!\r串列埠已關閉.", "例外狀況", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
        }

        /* 串列埠接收 ******************************************************* */

        private void serialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            int bytes = serialPort.BytesToRead;

            if (bytes > 0)
            {
                byte[] buffer = new byte[bytes];
                serialPort.Read(buffer, 0, bytes);

                for (int i = 0; i < bytes; i++)
                {
                    RX_Packet[RX_Index] = buffer[i];
                    if (RX_Index == 0)                          // Header 'S'
                    {
                        if (RX_Packet[0] == 0x53) RX_Index++;
                    }
                    else
                    {
                        if (RX_Index == 1)                      // Header 'M'
                        {
                            if (RX_Packet[1] == 0x4D) RX_Index++;
                            else RX_Index = 0;
                        }
                        else
                        {
                            if (RX_Index == 2)                  // 封包長度
                            {
                                RX_Length = RX_Packet[2];
                                if ((RX_Length == 5) || (RX_Length == 13)) RX_Index++;
                                else RX_Index = 0;
                            }
                            else
                            {
                                if (++RX_Index >= RX_Length)
                                {
                                    RX_Index = 0;

                                    ushort checksum = 0;
                                    byte n;
                                    for (n = 0; n < (RX_Length - 1); n++) checksum += RX_Packet[n];
                                    if ((checksum & 0xFF) == RX_Packet[n])
                                    {
                                        Packet_OK();            // 封包資料格式 OK
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        private void Packet_OK()                                // 封包資料格式 OK
        {
            switch (TX_Command)
            {
                case CMD_ReadData  : if (RX_Length == 13) FG_Receive = true;
                                     break;
                case CMD_L1_Set    :
                case CMD_L2_Set    :
                case CMD_L1_Output :
                case CMD_L2_Output :
                case CMD_L1_OnOff  :
                case CMD_L2_OnOff  :
                case CMD_L1_TrigOut:
                case CMD_L2_TrigOut: if (RX_Length == 5) FG_Receive = true;
                                     break;
            }
        }

        /* L1輸出調整 ******************************************************* */

        private void cbb_L1_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L1_Set >> 4) & 0x0F) != cbb_L1_MaxC.SelectedIndex)
            {
                byte i = (byte)(L1_Set & (SET_ExtOnOff | SET_OnOff));
                L1_Set = (byte)((cbb_L1_MaxC.SelectedIndex << 4) | i);

                nud_L1_Output.Value = 0;
                tkb_L1_Output.Value = 0;
                L1_Output = 0;
                CHG_L1_Output = false;

                int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
                lb_L1_MaxC.Text = (MaxC / 1000F).ToString("0.0");   // 以小數點顯示
                //tkb_L1_Output.TickFrequency = MaxC / 10;
                nud_L1_Output.Maximum = (decimal)(MaxC / 1000F);   // 以小數點顯示
                tkb_L1_Output.Maximum = MaxC;

                if (FG_ReadData == true) CHG_L1_Set = true;
            }
        }

        private void tkb_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            if (tkb_L1_Output.Value % 10 != 0) tkb_L1_Output.Value = (tkb_L1_Output.Value / 10 + 1) * 10;
            nud_L1_Output.Value = (decimal)(tkb_L1_Output.Value / 1000F);   // 以小數點顯示
            L1_Output = tkb_L1_Output.Value;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(L1_Output * 100) / MaxCurrent[cbb_L1_MaxC.SelectedIndex]);

            if (FG_ReadData == true) CHG_L1_Output = true;
        }

        private void nud_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L1_Output.Value = Convert.ToInt32(Math.Round(nud_L1_Output.Value * 1000, 0));
            L1_Output = tkb_L1_Output.Value;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(L1_Output * 100) / MaxCurrent[cbb_L1_MaxC.SelectedIndex]);

            if (FG_ReadData == true) CHG_L1_Output = true;
        }

        /* L1 ON/OFF ======================================================== */

        private void ckb_L1_ExtOnOff_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L1_ExtOnOff.Checked == false)               // Enable -> Disable ?
            {
                L1_Set &= (SET_ExtOnOff ^ 0xFF);                // L1 EXT. ON/OFF = Disable
                lb_L1_LED.Visible = true;
                btn_L1_OnOff.Visible = true;
            }
            else                                                // Disable -> Enable
            {
                L1_Set |= SET_ExtOnOff;                         // L1 EXT. ON/OFF = Enable
                lb_L1_LED.Visible = false;
                btn_L1_OnOff.Visible = false;
            }

            if (FG_ReadData == true) CHG_L1_Set = true;
        }

        private void btn_L1_OnOff_Click(object sender, EventArgs e)
        {
            if ((L1_Set & SET_OnOff) == 0)                      // OFF -> ON ?
            {
                L1_Set |= SET_OnOff;                            // L1 ON
                lb_L1_LED.BackColor = Color.Green;
                btn_L1_OnOff.Text = "OFF";
            }
            else                                                // ON -> OFF
            {
                L1_Set &= (SET_OnOff ^ 0xFF);                   // L1 OFF
                lb_L1_LED.BackColor = Color.DimGray;
                btn_L1_OnOff.Text = "ON";
            }

            if (FG_ReadData == true) CHG_L1_OnOff = true;
        }

        /* L1觸發輸出延遲 =================================================== */

        private void nud_L1_TrigOut_ValueChanged(object sender, EventArgs e)
        {
            L1_TrigOut = (byte)nud_L1_TrigOut.Value;
            lb_L1_TrigOut.Text = Convert.ToString(L1_TrigOut * 5) + "ms";

            if (FG_ReadData == true) CHG_L1_TrigOut = true;
        }

        /* L2輸出調整 ******************************************************* */

        private void cbb_L2_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L2_Set >> 4) & 0x0F) != cbb_L2_MaxC.SelectedIndex)
            {
                byte i = (byte)(L2_Set & (SET_ExtOnOff | SET_OnOff));
                L2_Set = (byte)((cbb_L2_MaxC.SelectedIndex << 4) | i);

                nud_L2_Output.Value = 0;
                tkb_L2_Output.Value = 0;
                L2_Output = 0;
                CHG_L2_Output = false;

                int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
                lb_L2_MaxC.Text = (MaxC / 1000F).ToString("0.0");   // 以小數點顯示
                //tkb_L2_Output.TickFrequency = MaxC / 10;
                nud_L2_Output.Maximum = (decimal)(MaxC / 1000F);   // 以小數點顯示
                tkb_L2_Output.Maximum = MaxC;

                if (FG_ReadData == true) CHG_L2_Set = true;
            }
        }

        private void tkb_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            if (tkb_L2_Output.Value % 10 != 0) tkb_L2_Output.Value = (tkb_L2_Output.Value / 10 + 1) * 10;
            nud_L2_Output.Value = (decimal)(tkb_L2_Output.Value / 1000F);   // 以小數點顯示
            L2_Output = tkb_L2_Output.Value;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(L2_Output * 100) / MaxCurrent[cbb_L2_MaxC.SelectedIndex]);

            if (FG_ReadData == true) CHG_L2_Output = true;
        }

        private void nud_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L2_Output.Value = Convert.ToInt32(Math.Round(nud_L2_Output.Value * 1000, 0));
            L2_Output = tkb_L2_Output.Value;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(L2_Output * 100) / MaxCurrent[cbb_L2_MaxC.SelectedIndex]);

            if (FG_ReadData == true) CHG_L2_Output = true;
        }

        /* L2 ON/OFF ======================================================== */

        private void ckb_L2_ExtOnOff_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L2_ExtOnOff.Checked == false)               // Enable -> Disable ?
            {
                L2_Set &= (SET_ExtOnOff ^ 0xFF);                // L2 EXT. ON/OFF = Disable
                lb_L2_LED.Visible = true;
                btn_L2_OnOff.Visible = true;
            }
            else                                                // Disable -> Enable
            {
                L2_Set |= SET_ExtOnOff;                         // L2 EXT. ON/OFF = Enable
                lb_L2_LED.Visible = false;
                btn_L2_OnOff.Visible = false;
            }

            if (FG_ReadData == true) CHG_L2_Set = true;
        }

        private void btn_L2_OnOff_Click(object sender, EventArgs e)
        {
            if ((L2_Set & SET_OnOff) == 0)                      // OFF -> ON
            {
                L2_Set |= SET_OnOff;                            // L2 ON
                lb_L2_LED.BackColor = Color.Green;
                btn_L2_OnOff.Text = "OFF";
            }
            else                                                // ON -> OFF
            {
                L2_Set &= (SET_OnOff ^ 0xFF);                   // L2 OFF
                lb_L2_LED.BackColor = Color.DimGray;
                btn_L2_OnOff.Text = "ON";
            }

            if (FG_ReadData == true) CHG_L2_OnOff = true;
        }

        /* L2觸發輸出延遲 =================================================== */

        private void nud_L2_TrigOut_ValueChanged(object sender, EventArgs e)
        {
            L2_TrigOut = (byte)nud_L2_TrigOut.Value;
            lb_L2_TrigOut.Text = Convert.ToString(L2_TrigOut * 5) + "ms";

            if (FG_ReadData == true) CHG_L2_TrigOut = true;
        }
    }
}
